# Copyright (C) 2017 Hans van Kranenburg <hans@knorrie.org>
#
# This file is part of the python-btrfs module.
#
# python-btrfs is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-btrfs is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-btrfs.  If not, see <http://www.gnu.org/licenses/>.

table = (
    0x00000000, 0xf26b8303, 0xe13b70f7, 0x1350f3f4,
    0xc79a971f, 0x35f1141c, 0x26a1e7e8, 0xd4ca64eb,
    0x8ad958cf, 0x78b2dbcc, 0x6be22838, 0x9989ab3b,
    0x4d43cfd0, 0xbf284cd3, 0xac78bf27, 0x5e133c24,
    0x105ec76f, 0xe235446c, 0xf165b798, 0x030e349b,
    0xd7c45070, 0x25afd373, 0x36ff2087, 0xc494a384,
    0x9a879fa0, 0x68ec1ca3, 0x7bbcef57, 0x89d76c54,
    0x5d1d08bf, 0xaf768bbc, 0xbc267848, 0x4e4dfb4b,
    0x20bd8ede, 0xd2d60ddd, 0xc186fe29, 0x33ed7d2a,
    0xe72719c1, 0x154c9ac2, 0x061c6936, 0xf477ea35,
    0xaa64d611, 0x580f5512, 0x4b5fa6e6, 0xb93425e5,
    0x6dfe410e, 0x9f95c20d, 0x8cc531f9, 0x7eaeb2fa,
    0x30e349b1, 0xc288cab2, 0xd1d83946, 0x23b3ba45,
    0xf779deae, 0x05125dad, 0x1642ae59, 0xe4292d5a,
    0xba3a117e, 0x4851927d, 0x5b016189, 0xa96ae28a,
    0x7da08661, 0x8fcb0562, 0x9c9bf696, 0x6ef07595,
    0x417b1dbc, 0xb3109ebf, 0xa0406d4b, 0x522bee48,
    0x86e18aa3, 0x748a09a0, 0x67dafa54, 0x95b17957,
    0xcba24573, 0x39c9c670, 0x2a993584, 0xd8f2b687,
    0x0c38d26c, 0xfe53516f, 0xed03a29b, 0x1f682198,
    0x5125dad3, 0xa34e59d0, 0xb01eaa24, 0x42752927,
    0x96bf4dcc, 0x64d4cecf, 0x77843d3b, 0x85efbe38,
    0xdbfc821c, 0x2997011f, 0x3ac7f2eb, 0xc8ac71e8,
    0x1c661503, 0xee0d9600, 0xfd5d65f4, 0x0f36e6f7,
    0x61c69362, 0x93ad1061, 0x80fde395, 0x72966096,
    0xa65c047d, 0x5437877e, 0x4767748a, 0xb50cf789,
    0xeb1fcbad, 0x197448ae, 0x0a24bb5a, 0xf84f3859,
    0x2c855cb2, 0xdeeedfb1, 0xcdbe2c45, 0x3fd5af46,
    0x7198540d, 0x83f3d70e, 0x90a324fa, 0x62c8a7f9,
    0xb602c312, 0x44694011, 0x5739b3e5, 0xa55230e6,
    0xfb410cc2, 0x092a8fc1, 0x1a7a7c35, 0xe811ff36,
    0x3cdb9bdd, 0xceb018de, 0xdde0eb2a, 0x2f8b6829,
    0x82f63b78, 0x709db87b, 0x63cd4b8f, 0x91a6c88c,
    0x456cac67, 0xb7072f64, 0xa457dc90, 0x563c5f93,
    0x082f63b7, 0xfa44e0b4, 0xe9141340, 0x1b7f9043,
    0xcfb5f4a8, 0x3dde77ab, 0x2e8e845f, 0xdce5075c,
    0x92a8fc17, 0x60c37f14, 0x73938ce0, 0x81f80fe3,
    0x55326b08, 0xa759e80b, 0xb4091bff, 0x466298fc,
    0x1871a4d8, 0xea1a27db, 0xf94ad42f, 0x0b21572c,
    0xdfeb33c7, 0x2d80b0c4, 0x3ed04330, 0xccbbc033,
    0xa24bb5a6, 0x502036a5, 0x4370c551, 0xb11b4652,
    0x65d122b9, 0x97baa1ba, 0x84ea524e, 0x7681d14d,
    0x2892ed69, 0xdaf96e6a, 0xc9a99d9e, 0x3bc21e9d,
    0xef087a76, 0x1d63f975, 0x0e330a81, 0xfc588982,
    0xb21572c9, 0x407ef1ca, 0x532e023e, 0xa145813d,
    0x758fe5d6, 0x87e466d5, 0x94b49521, 0x66df1622,
    0x38cc2a06, 0xcaa7a905, 0xd9f75af1, 0x2b9cd9f2,
    0xff56bd19, 0x0d3d3e1a, 0x1e6dcdee, 0xec064eed,
    0xc38d26c4, 0x31e6a5c7, 0x22b65633, 0xd0ddd530,
    0x0417b1db, 0xf67c32d8, 0xe52cc12c, 0x1747422f,
    0x49547e0b, 0xbb3ffd08, 0xa86f0efc, 0x5a048dff,
    0x8ecee914, 0x7ca56a17, 0x6ff599e3, 0x9d9e1ae0,
    0xd3d3e1ab, 0x21b862a8, 0x32e8915c, 0xc083125f,
    0x144976b4, 0xe622f5b7, 0xf5720643, 0x07198540,
    0x590ab964, 0xab613a67, 0xb831c993, 0x4a5a4a90,
    0x9e902e7b, 0x6cfbad78, 0x7fab5e8c, 0x8dc0dd8f,
    0xe330a81a, 0x115b2b19, 0x020bd8ed, 0xf0605bee,
    0x24aa3f05, 0xd6c1bc06, 0xc5914ff2, 0x37faccf1,
    0x69e9f0d5, 0x9b8273d6, 0x88d28022, 0x7ab90321,
    0xae7367ca, 0x5c18e4c9, 0x4f48173d, 0xbd23943e,
    0xf36e6f75, 0x0105ec76, 0x12551f82, 0xe03e9c81,
    0x34f4f86a, 0xc69f7b69, 0xd5cf889d, 0x27a40b9e,
    0x79b737ba, 0x8bdcb4b9, 0x988c474d, 0x6ae7c44e,
    0xbe2da0a5, 0x4c4623a6, 0x5f16d052, 0xad7d5351,
)


def crc32c(crc, data):
    """
    General crc32c function. You likely don't need to call this one, but use
    one of the more specific helpers in this module.

    :param int crc: Seed value.
    :param bytes data: Bytes to checksum.
    :returns: A 32 bit checksum.
    :rtype: int
    """
    if not isinstance(data, (bytes, bytearray)):
        data = bytes(data, 'utf-8')
    crc = crc & 0xffffffff
    for char in data:
        crc = table[(crc ^ char) & 0xff] ^ (crc >> 8)
    return crc & 0xffffffff


def crc32c_data(data):
    """
    Compute the checksum for a sectorsize block of bytes. These are the kind of
    checksum that go into the CSUM tree.

    :param bytes data: A single disk block of data.
    :returns: Btrfs checksum for the data.
    :rtype: int
    """
    return 0xffffffff ^ crc32c(~0, data)


def name_hash(name):
    """
    The name hash function is used to compute a crc32c of a filename. The
    resulting value is used in the offset field of they key of a
    :class:`~btrfs.ctree.DirItem` object in subvolume trees.

    By searching for the numeric value, a file with a certain name can quickly
    be found without searching the whole directory content.

    :param bytes name: File name as bytes. For convenience, if a unicode string
        is provided, it will be encoded as utf-8.
    :returns: A btrfs name hash.
    :rtype: int

    Example::

        >>> btrfs.crc32c.name_hash('mouton')
        3786996654
    """
    return crc32c(~1, name)


def extref_hash(parent_objectid, name):
    """
    The extref_hash function uses the crc32c code with the inode number of the
    containing directory, and a filename to compute the offset field for the
    key of an :class:`~btrfs.ctree.InodeExtref` objects.

    :param int parent_objectid: ObjectID of the containing directory.
    :param bytes name: File name as bytes. For convenience, if a unicode string
        is provided, it will be encoded as utf-8.
    :returns: A btrfs extref hash.
    :rtype: int

    Example::

        >>> btrfs.crc32c.extref_hash(256, 'mouton')
        1903957879
    """
    return crc32c(parent_objectid, name)



# Copyright (C) 2019 Hans van Kranenburg <hans@knorrie.org>
#
# This file is part of the python-btrfs module.
#
# python-btrfs is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-btrfs is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-btrfs.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains the library version number. This must be the only place
where the version number is kept.
"""

__version__ = '15'



# Copyright (C) 2018 Hans van Kranenburg <hans@knorrie.org>
#
# This file is part of the python-btrfs module.
#
# python-btrfs is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-btrfs is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-btrfs.  If not, see <http://www.gnu.org/licenses/>.

"""
This module provides advanced usage reporting for a btrfs filesystem.

By calling the :func:`~btrfs.ctree.FileSystem.usage` function on a
:class:`btrfs.ctree.FileSystem` object, an :class:`FsUsage` object is returned
that can be inspected.

Example::

    >>> import btrfs
    >>> with btrfs.FileSystem('/') as fs:
    ...     usage = fs.usage()
    ...     btrfs.utils.pretty_print(usage)

.. note::

    If you're not yet familiar with it, btrfs terminology can be quite
    confusing.

    Here's just an example: In btrfs terminology, a ‘space’ is the collection
    of all block groups that have identical type and profile flags. For
    example, Metadata, DUP is a ‘space’. The word ‘space’ is also used for the
    distinction between ‘physical address space’ and ‘virtual address space’.

"""

import btrfs
import copy
from btrfs.ctree import (  # noqa
    BLOCK_GROUP_DATA, BLOCK_GROUP_SYSTEM, BLOCK_GROUP_METADATA,
    BLOCK_GROUP_TYPE_MASK, BLOCK_GROUP_PROFILE_MASK,
)


class DevSpaceUsage(object):
    """Physical usage details for a single space per device.

    For example, a `Data, DUP` chunk of 1GiB results in a 2GiB allocation of
    physical bytes on the device. A `Data, RAID5` chunk of 3GiB, allocated over
    4 devices results in a 1GiB allocation on each device, with 256MiB reserved
    for parity.

    :ivar int flags: Block group type and profile, e.g. `Data, RAID1`.
    :ivar int devid: Device ID
    :ivar int allocated: Total amount of allocated physical bytes.
    :ivar int parity: Amount of allocated physical bytes reserved for parity.

    .. note::
        Objects of this type are provided as part of an :class:`FsUsage` object.

    """
    def __init__(self, devid, flags):
        self.flags = flags
        self.devid = devid
        self.allocated = 0
        self.parity = 0

    def _add_usage(self, allocated, parity):
        self.allocated += allocated
        self.parity += parity

    @staticmethod
    def _pretty_properties():
        return [
            (btrfs.utils.block_group_flags_str, 'flags'),
            (btrfs.utils.pretty_size, 'allocated'),
            (btrfs.utils.pretty_size, 'parity'),
        ]


class DevUsage(object):
    """Physical usage details for a device.

    :ivar int devid: Device ID
    :ivar int total: Total amount of bytes.
    :ivar int allocated: Total amount of allocated bytes.
    :ivar dev_space_usage: Allocated and parity bytes per space for this
        device, indexed by space flags.
    :vartype dev_space_usage: dict of DevSpaceUsage
    :ivar int unallocatable: Physical bytes that are not allocatable because of
        unbalanced device sizes.
    :ivar int unallocatable_reclaimable: Physical bytes that are not
        allocatable because of unbalanced allocations.

    .. note::
        Objects of this type are provided as part of an :class:`FsUsage` object.

    """
    def __init__(self, device):
        self.devid = device.devid
        self.total = device.total_bytes
        self.allocated = device.bytes_used
        self.unallocated = self.total - self.allocated
        self.dev_space_usage = {}
        self.unallocatable_soft = None  # set during FsUsage init
        self.unallocatable_hard = None  # set during FsUsage init
        self.unallocatable_reclaimable = None  # set during FsUsage init

    def _dev_space_usage_key_str(flags):
        return btrfs.utils.block_group_flags_str(flags)

    def _add_usage(self, flags, allocated, parity):
        if flags not in self.dev_space_usage:
            self.dev_space_usage[flags] = DevSpaceUsage(self.devid, flags)
        self.dev_space_usage[flags]._add_usage(allocated, parity)

    def _init_unallocatable_reclaimable(self):
        self.unallocatable_reclaimable = \
            max(self.unallocatable_soft - self.unallocatable_hard, 0)

    @staticmethod
    def _pretty_properties():
        return [
            (btrfs.utils.pretty_size, 'total'),
            (btrfs.utils.pretty_size, 'allocated'),
            (btrfs.utils.pretty_size, 'unallocated'),
            (btrfs.utils.pretty_size, 'unallocatable_soft'),
            (btrfs.utils.pretty_size, 'unallocatable_hard'),
            (btrfs.utils.pretty_size, 'unallocatable_reclaimable'),
        ]


class RawSpaceUsage(object):
    """Physical usage details per space.

    For example, if the `Metadata, RAID1` space has a 2GiB size in terms of
    virtual addressing, in which 768MiB is used, then the allocated physical
    size is 4GiB and amount of physical bytes used is 1.5GiB. A `Data, RAID6`
    space of 8GiB, consisting of two 4GiB block groups (virtual address space),
    each distributed over 6 devices, will occupy 2*(4+2) = 12 GiB physical
    allocated bytes and have 4GiB of allocated bytes reserved for parity
    blocks.

    :ivar int flags: Block group type and profile, e.g. `Data, RAID1`.
    :ivar int allocated: Total amount of allocated bytes.
    :ivar int parity: Total amount of allocated bytes reserved for parity
        blocks.
    :ivar int used: Total amount of physical bytes used.

    .. note::
        Objects of this type are provided as part of an :class:`FsUsage` object.

    """
    def __init__(self, space):
        self.flags = space.flags
        ratio = btrfs.volumes.block_group_profile_ncopies(space.flags)
        self.allocated = space.total_bytes * ratio  # initially missing parity blocks
        self.used = space.used_bytes * ratio
        self.parity = 0

    def _add_usage(self, parity):
        self.allocated += parity
        self.parity += parity

    @staticmethod
    def _pretty_properties():
        return [
            (btrfs.utils.block_group_flags_str, 'flags'),
            (btrfs.utils.pretty_size, 'allocated'),
            (btrfs.utils.pretty_size, 'parity'),
            (btrfs.utils.pretty_size, 'used'),
        ]


class BlockGroupTypeUsage(object):
    """Physical usage details per block group type.

    Totals per block group type (`System`, `Data`, `Metadata`, or,
    `Data+Metadata` for mixed mode), disregarding the block group profile
    (`Single`, `RAID1`, etc).

    :ivar int type: Block group type, e.g. `Metadata`.
    :ivar int allocated: Total amount of allocated bytes.
    :ivar int parity: Total amount of allocated bytes reserved for parity
        blocks.
    :ivar int used: Total amount of physical bytes used.

    .. note::
        Objects of this type are provided as part of an :class:`FsUsage` object.

    """
    def __init__(self, block_group_type):
        self.type = block_group_type
        self.allocated = 0
        self.used = 0
        self.parity = 0

    def _add_usage(self, allocated, parity, used):
        self.allocated += allocated
        self.parity += parity
        self.used += used

    @staticmethod
    def _pretty_properties():
        return [
            (btrfs.utils.block_group_flags_str, 'type'),
            (btrfs.utils.pretty_size, 'allocated'),
            (btrfs.utils.pretty_size, 'parity'),
            (btrfs.utils.pretty_size, 'used'),
        ]


class VirtualSpaceUsage(object):
    """Virtual usage per space.

    :ivar int flags: Block group type and profile, e.g. `Data, RAID1`.
    :ivar int total: Total amount of allocated bytes for this space.
    :ivar int used: Total amount of virtual bytes used.

    .. note::
        Objects of this type are provided as part of an :class:`FsUsage` object.

    """
    def __init__(self, space):
        self.flags = space.flags
        self.total = space.total_bytes
        self.used = space.used_bytes

    @staticmethod
    def _pretty_properties():
        return [
            (btrfs.utils.block_group_flags_str, 'flags'),
            (btrfs.utils.pretty_size, 'total'),
            (btrfs.utils.pretty_size, 'used'),
        ]


class VirtualBlockGroupTypeUsage(object):
    """Virtual address space usage per block group type.

    Totals for the virtual address space per block group type (`System`,
    `Data`, `Metadata`, or, `Data+Metadata` for mixed mode), disregarding the
    block group profile (`Single`, `RAID1`, etc).

    :ivar int flags: Block group type, e.g. `Metadata`.
    :ivar int total: Total amount of allocated bytes.
    :ivar int used: Total amount of virtual bytes used.
    :ivar int unused: Amount of allocated but unused virtual bytes.

    .. note::
        Objects of this type are provided as part of an :class:`FsUsage` object.

    """
    def __init__(self, block_group_type):
        self.type = block_group_type
        self.total = 0
        self.used = 0
        self.unused = None

    def _add_usage(self, total, used):
        self.total += total
        self.used += used
        self.unused = self.total - self.used

    @staticmethod
    def _pretty_properties():
        return [
            (btrfs.utils.block_group_flags_str, 'type'),
            (btrfs.utils.pretty_size, 'total'),
            (btrfs.utils.pretty_size, 'used'),
            (btrfs.utils.pretty_size, 'unused'),
        ]


class FsUsage(object):
    """Detailed usage information for a file system.

    When creating an object of this type, the first argument, fs, is mandatory.
    The other arguments can be used to influence the simulation to predict free
    space and unallocatable space with explicit hints instead of using
    information from the current filesystem, This is used by the
    space-calculator program to run the simulation starting with a completely
    empty filesystem.

    :param btrfs.ctree.FileSystem fs: Filesystem to examine.
    :param int data_metadata_ratio: Data to metadata ratio to use when running
        the simulation to predict free space and unallocatable space.
    :param int target_profile_metadata: Explicitly set metadata profile to use
        for new allocations when running the simulation to predict free
        and unallocatable space (not for a mixed filesystem).
    :param int target_profile_data: Explicitly set data profile to use for new
        allocations when running the simulation to predict free and
        unallocatable space (not for a mixed filesystem).
    :param int target_profile_mixed: Explicitly set metadata and data profile
        to use for new allocations when running the simulation to predict free
        and unallocatable space (only for a mixed filesystem).

    Target block group profiles (used for new chunk allocations):

    :ivar int target_profile_system: Profile for new System chunk allocations.
    :ivar int target_profile_metadata: Profile for new Metadata chunk
        allocations (not for a mixed filesystem).
    :ivar int target_profile_data: Profile for new Data chunk allocations (not
        for a mixed filesystem).
    :ivar int target_profile_mixed: Profile for new Data+Metadata chunk
        allocations (only for a mixed filesystem).

    Usage details for the physical address space:

    :ivar int total: Total amount of physical bytes in the filesystem.
    :ivar int allocated: Total amount of allocated physical bytes.
    :ivar int parity: Total amount of allocated bytes reserved for parity
        blocks.
    :ivar dev_usage: Physical usage details per device, indexed by Device ID.
    :vartype dev_usage: dict of DevUsage
    :ivar block_group_type_usage: Physical usage details per block group type,
        indexed by block group type.
    :vartype block_group_type_usage: dict of BlockGroupTypeUsage
    :ivar raw_space_usage: Physical usage details per space, indexed by space
        flags.
    :vartype raw_space_usage: dict of RawSpaceUsage

    Usage details for the virtual address space:

    :ivar int virtual_total: Total amount of virtual address space.
    :ivar int virtual_used: Total amount of bytes used inside the virtual
        address space.
    :ivar virtual_block_group_type_usage: Virtual address space usage per block
        group type, indexed by block group type.
    :vartype virtual_block_group_type_usage: dict of VirtualBlockGroupTypeUsage
    :ivar VirtualSpaceUsage virtual_space_usage: Virtual usage per space,
        indexed by space flags.
    :vartype virtual_space_usage: dict of VirtualSpaceUsage

    Allocatable space information:

    The *soft* unallocatable amount of bytes is the currently unallocatable
    part of the physical bytes on attached devices because the allocations in
    the filesystem are unbalanced. This value is estimated by extrapolating the
    current usage pattern and simulating new chunk allocations using the
    current target allocation profiles. By doing so, we also discover how much
    extra *virtual* address space these allocations would result in.

    :ivar int unallocatable_soft: Unallocatable physical disk space because of
        unbalanced allocations.
    :ivar int estimated_allocatable_virtual_metadata: Estimated amount of
        virtual address space bytes that can be added by allocating physical
        bytes for metadata, based on the current usage pattern (not for a mixed
        filesystem).
    :ivar int estimated_allocatable_virtual_data: Estimated amount of virtual
        address space bytes that can be added by allocating physical bytes for
        data, based on the current usage pattern (not for a mixed filesystem).
    :ivar int estimated_allocatable_virtual_mixed: Estimated amount of virtual
        address space bytes that can be added by allocating physical bytes for
        metadata and data,, based on the current usage pattern (only for a
        mixed filesystem).

    The *hard* unallocatable amount of bytes is the amount of physical bytes
    that cannot be used for allocations, because of having different sizes of
    devices attached. These values are unallocatable disk space that remains
    after trying to simulate data and metadata allocations in a ratio similar
    to current usage, starting with all disks being empty.

    :ivar int unallocatable_hard: Unallocatable physical disk space because of
        unbalanced device sizes.
    :ivar int estimated_full_allocatable_virtual_metadata: Estimated amount of
        virtual address space bytes for metadata, in case of optimally balanced
        allocations. (not for a mixed filesystem).
    :ivar int estimated_full_allocatable_virtual_data: Estimated amount of
        virtual address space bytes for data, in case of optimally balanced
        allocations. (not for a mixed filesystem).
    :ivar int estimated_full_allocatable_virtual_mixed: Estimated amount of
        virtual address space bytes for metadata and data, in case of optimally
        balanced allocations. (only for a mixed filesystem).

    The difference between *soft* and *hard* unallocatable bytes is the amount
    of physical disk space that can be reclaimed for allocations when
    rebalancing the filesystem.

    :ivar int unallocatable_reclaimable: Unallocatable physical bytes that can
        be reclaimed when balancing the filesystem.

    Some other totals for convenience:

    :ivar int allocatable: The total amount of physical bytes that are
        allocatable in this filesystem. I.e. total size minus
        unallocatable_soft.
    :ivar int allocatable_left: The amount of allocatable physical bytes
        remaining, until the filesystem will report being out of space, given
        current usage pattern and target profiles.

    By combining the unused virtual space in already allocated chunks and
    estimated allocatable virtual bytes, we get actual numbers of estimated
    free space. I.e., what we would like df to show us.

    :ivar int free_metadata: Estimated virtual space left to use for metadata
        (not for a mixed filesystem).
    :ivar int free_data: Estimated virtual space left to use for data (not for
        a mixed filesystem).
    :ivar int free_mixed: Estimated virtual space left to use for metadata and
        data (only for a mixed filesystem).

    """
    def __init__(self, fs, data_metadata_ratio=None,
                 target_profile_metadata=None,
                 target_profile_data=None,
                 target_profile_mixed=None):
        self._mixed_groups = fs.mixed_groups()

        # Spaces and devices are a source of information
        spaces = [
            space
            for space in fs.space_info()
            if space.flags != btrfs.ctree.SPACE_INFO_GLOBAL_RSV
        ]
        devices = list(fs.devices())

        self.raw_space_usage = {
            space.flags: RawSpaceUsage(space)
            for space in spaces
        }
        self.dev_usage = {
            device.devid: DevUsage(device)
            for device in devices
        }
        self.total = sum(
            device.total_bytes
            for device in devices
        )
        self.allocated = sum(
            device.bytes_used
            for device in devices
        )
        self.parity = 0

        BLOCK_GROUP_MIXED = btrfs.BLOCK_GROUP_METADATA | btrfs.BLOCK_GROUP_DATA
        if not self._mixed_groups:
            if target_profile_metadata is not None:
                self.target_profile_system = target_profile_metadata | btrfs.BLOCK_GROUP_SYSTEM
                self.target_profile_metadata = target_profile_metadata | btrfs.BLOCK_GROUP_METADATA
            if target_profile_data is not None:
                self.target_profile_data = target_profile_data | btrfs.BLOCK_GROUP_DATA
        else:
            if target_profile_mixed is not None:
                self.target_profile_system = target_profile_mixed | btrfs.BLOCK_GROUP_SYSTEM
                self.target_profile_mixed = target_profile_mixed | BLOCK_GROUP_MIXED

        # We walk the chunk list because every block group / chunk can be laid
        # out over any amount of disks. To collect e.g. the amount of parity
        # bytes, we need to look at all of them.
        #
        # Confusing: chunk.type is actually all the flags, so type and profile
        # combined.
        for chunk in fs.chunks():
            flags = chunk.type
            if flags not in self.raw_space_usage:
                continue  # A conversion to this profile just started right now?
            # Remember last seen chunk types as target profiles
            block_group_type = flags & BLOCK_GROUP_TYPE_MASK
            if block_group_type == BLOCK_GROUP_SYSTEM:
                self.target_profile_system = flags
            elif not self._mixed_groups:
                if block_group_type == BLOCK_GROUP_DATA:
                    self.target_profile_data = flags
                elif block_group_type == BLOCK_GROUP_METADATA:
                    self.target_profile_metadata = flags
            else:
                self.target_profile_mixed = flags

            dev_extent_length = btrfs.volumes.chunk_to_dev_extent_length(chunk)
            chunk_raw_parity_bytes = btrfs.volumes.chunk_to_raw_parity_bytes(chunk)
            self.parity += chunk_raw_parity_bytes
            self.raw_space_usage[flags]._add_usage(chunk_raw_parity_bytes)
            dev_extent_parity_bytes = chunk_raw_parity_bytes // len(chunk.stripes)
            for stripe in chunk.stripes:
                if stripe.devid not in self.dev_usage:
                    continue  # A device just got added?
                self.dev_usage[stripe.devid]._add_usage(flags,
                                                        dev_extent_length,
                                                        dev_extent_parity_bytes)

        # Combine information from different spaces with same chunk type into
        # totals per block group type. So, e.g. all DATA space, regardless of
        # being single, RAID1, etc...
        self.block_group_type_usage = {}
        for raw_space in self.raw_space_usage.values():
            space_type = raw_space.flags & BLOCK_GROUP_TYPE_MASK
            if space_type not in self.block_group_type_usage:
                self.block_group_type_usage[space_type] = BlockGroupTypeUsage(space_type)
            self.block_group_type_usage[space_type]._add_usage(
                    raw_space.allocated, raw_space.parity, raw_space.used)

        self.virtual_space_usage = {
            space.flags: VirtualSpaceUsage(space)
            for space in spaces
        }

        # Combine information from different spaces with same chunk type into
        # totals per block group type. So, e.g. all DATA space, regardless of
        # being single, RAID1, etc...
        self.virtual_block_group_type_usage = {}

        for virtual_space in self.virtual_space_usage.values():
            space_type = virtual_space.flags & BLOCK_GROUP_TYPE_MASK
            if space_type not in self.virtual_block_group_type_usage:
                self.virtual_block_group_type_usage[space_type] = \
                        VirtualBlockGroupTypeUsage(space_type)
            self.virtual_block_group_type_usage[space_type]._add_usage(
                    virtual_space.total, virtual_space.used)

        # The total size of the block groups in the virtual address space
        self.virtual_total = sum(
            virtual_space.total
            for virtual_space in self.virtual_space_usage.values()
        )

        self.virtual_used = sum(
            virtual_space.used
            for virtual_space in self.virtual_space_usage.values()
        )

        if data_metadata_ratio is not None:
            self.default_data_metadata_ratio = data_metadata_ratio
        else:
            self.default_data_metadata_ratio = 200

        # Estimate the amount of unallocatable raw disk space if the sizes of
        # attached block devices are unbalanced. We start the simulation with
        # the entire sizes of the attached devices and keep allocating chunks
        # until not possible any more.
        #
        device_sizes = {
            devid: dev_usage.total
            for devid, dev_usage in self.dev_usage.items()
        }
        if not self._mixed_groups:
            # The estimated "full" allocatable numbers are the estimation of
            # virtual space to be used for data and metadata when the
            # filesystem would be totally empty to begin with.
            #
            # This is e.g. used by the space calculator example.
            dev_unallocatable_hard, \
                self.estimated_full_allocatable_virtual_metadata, \
                self.estimated_full_allocatable_virtual_data = \
                self._simulate_chunk_allocations(device_sizes)
        else:
            dev_unallocatable_hard, \
                self.estimated_full_allocatable_virtual_mixed = \
                self._simulate_chunk_allocations(device_sizes)
        for devid, unallocatable_hard in dev_unallocatable_hard.items():
            self.dev_usage[devid].unallocatable_hard = unallocatable_hard
        self.unallocatable_hard = sum(dev_unallocatable_hard.values())

        # Next, we estimate the amount of unallocatable raw disk space when
        # starting out with the current state of the filesystem.
        unallocated_sizes = {
            devid: dev_usage.unallocated
            for devid, dev_usage in self.dev_usage.items()
        }
        if not self._mixed_groups:
            dev_unallocatable_soft, \
                self.estimated_allocatable_virtual_metadata, \
                self.estimated_allocatable_virtual_data = \
                self._simulate_chunk_allocations(unallocated_sizes)
        else:
            dev_unallocatable_soft, \
                self.estimated_allocatable_virtual_mixed = \
                self._simulate_chunk_allocations(unallocated_sizes)
        for devid, unallocatable_soft in dev_unallocatable_soft.items():
            self.dev_usage[devid].unallocatable_soft = unallocatable_soft
        self.unallocatable_soft = sum(dev_unallocatable_soft.values())

        # At this point, it is possible that the unallocatable_hard amounts are
        # higher than the unallocatable_soft amounts. E.g. if we just switched
        # to another target profile. In that case, this means we can not fully
        # rewrite all data to the target profile!

        # For convenience reasons, we provide a few more derived numbers...
        #
        # If the soft unallocatable amount of bytes is higher than the hard
        # amount, we can reclaim space by balancing the filesystem.
        self.unallocatable_reclaimable = \
            max(self.unallocatable_soft - self.unallocatable_hard, 0)
        for dev_usage in self.dev_usage.values():
            dev_usage._init_unallocatable_reclaimable()

        self.allocatable = self.total - self.unallocatable_soft
        self.allocatable_left = self.allocatable - self.allocated

        if not self._mixed_groups:
            self.free_metadata = self.estimated_allocatable_virtual_metadata
            if btrfs.BLOCK_GROUP_METADATA in self.virtual_block_group_type_usage:
                self.free_metadata += \
                    self.virtual_block_group_type_usage[btrfs.BLOCK_GROUP_METADATA].unused
            self.free_data = self.estimated_allocatable_virtual_data
            if btrfs.BLOCK_GROUP_DATA in self.virtual_block_group_type_usage:
                self.free_data += \
                    self.virtual_block_group_type_usage[btrfs.BLOCK_GROUP_DATA].unused
        else:
            self.free_mixed = self.estimated_allocatable_virtual_mixed
            if BLOCK_GROUP_MIXED in self.virtual_block_group_type_usage:
                self.free_mixed += \
                    self.virtual_block_group_type_usage[BLOCK_GROUP_MIXED].unused

    def _raw_space_usage_key_str(flags):
        return btrfs.utils.block_group_flags_str(flags)

    def _block_group_type_usage_key_str(block_group_type):
        return btrfs.utils.block_group_flags_str(block_group_type)

    def _virtual_space_usage_key_str(flags):
        return btrfs.utils.block_group_flags_str(flags)

    def _virtual_block_group_type_usage_key_str(block_group_type):
        return btrfs.utils.block_group_flags_str(block_group_type)

    def _data_metadata_ratio(self):
        """
        Determine data to metadata ratio for allocation simulation for wasted space.

        E.g. a ratio of 200 means that for every X bytes of metadata, we can
        allocate 200*X bytes for data.

        Actual simulation ratio is a weighted combination of the current data to
        metadata ratio and the default of 200, where the weight of the current
        ratio increases if the filesystem is filled up more.

        If an empty filesystem is presented (this is used for the space
        calculator), then just use the default that was set earlier.
        """
        if self._mixed_groups:
            raise ValueError("Data to metadata ratio is irrelevant for mixed groups.")
        if BLOCK_GROUP_METADATA not in self.virtual_block_group_type_usage:
            return self.default_data_metadata_ratio
        used_fraction = self.virtual_used / self.total
        used_metadata = self.virtual_block_group_type_usage[BLOCK_GROUP_METADATA].used
        used_data = self.virtual_block_group_type_usage[BLOCK_GROUP_DATA].used
        used_ratio = used_data / used_metadata
        return used_fraction * used_ratio + (1 - used_fraction) * self.default_data_metadata_ratio

    def _alloc_chunk(self, sizes, flags):
        """
        This is used by the wasted space calculator.

        sizes is a dictionary {devid: allocatable_bytes, ...}
              which will be modified in place as side effect
        flags contains allocation type, which is DATA (also for mixed) or METADATA

        This function tries to reduce unallocated raw bytes on each disk in a way
        similar to the workings of the btrfs chunk allocator. The sizes dictionary
        will be modified in place while doing so. When returning False, no chunk
        allocation is possible any more, and sizes will show the amount of
        unallocatable bytes per device.
        """
        attrs = btrfs.volumes._raid_attrs(flags & BLOCK_GROUP_PROFILE_MASK)
        if flags & BLOCK_GROUP_DATA:
            max_stripe_size = btrfs.utils.SZ_1G
            max_chunk_size = btrfs.volumes.BTRFS_MAX_DATA_CHUNK_SIZE
        elif flags & BLOCK_GROUP_METADATA:
            if self.total > 50 * btrfs.utils.SZ_1G:
                max_stripe_size = btrfs.utils.SZ_1G
            else:
                max_stripe_size = btrfs.utils.SZ_256M
            max_chunk_size = max_stripe_size
        else:
            raise ValueError("Only DATA and METADATA supported here")
        # we don't want a chunk larger than 10% of writeable space
        max_chunk_size = min(self.total // 10, max_chunk_size)
        # [(devid, unallocated), ...], most unallocated space per device first
        non_zero_sizes = {
            devid: unallocated
            for devid, unallocated in sizes.items()
            if unallocated > 0
        }
        sorted_sizes = sorted(list(non_zero_sizes.items()), key=lambda x: -x[1])
        # Keep a multiple of devs_increment, chop off the rest
        sorted_sizes = sorted_sizes[:len(sorted_sizes) -
                                    (len(sorted_sizes) % attrs.devs_increment)]
        if len(sorted_sizes) < attrs.devs_min:
            return 0
        # Keep only the amount we need for a single chunk allocation
        if attrs.devs_max != 0:
            sorted_sizes = sorted_sizes[:min(len(sorted_sizes), attrs.devs_max)]
        # Actual device extent size is limited by the device with least amount of
        # available space and by max_stripe_size.
        stripe_size = min(max_stripe_size, sorted_sizes[-1][1] // attrs.dev_stripes)
        # But, there's another limit, the max_chunk_size...
        num_stripes = len(sorted_sizes) * attrs.dev_stripes
        chunk_size = btrfs.volumes.dev_extent_length_to_chunk_length(
                flags, num_stripes, stripe_size)
        if chunk_size > max_chunk_size:
            stripe_size = btrfs.volumes.chunk_length_to_dev_extent_length(
                    flags, num_stripes, max_chunk_size)
            chunk_size = max_chunk_size
        # Finally, decrease unallocated space
        for devid, _ in sorted_sizes:
            sizes[devid] -= stripe_size * attrs.dev_stripes
        return chunk_size

    def _simulate_chunk_allocations(self, sizes):
        """
        Try to do metadata and data allocations until no longer possible. The sizes
        dictionary is modified in place and lists the amount of unallocatable space
        per disk when the function returns.
        """
        _sizes = copy.deepcopy(sizes)  # copy will be modified in place
        if not self._mixed_groups:
            ratio = self._data_metadata_ratio()
            metadata_flags = self.target_profile_metadata
            data_flags = self.target_profile_data
            virtual_data = 0
            virtual_metadata = 0
            while True:
                chunk_size = self._alloc_chunk(_sizes, metadata_flags)
                if chunk_size == 0:
                    return _sizes, virtual_metadata, virtual_data
                virtual_metadata += chunk_size
                while virtual_data / virtual_metadata < ratio:
                    chunk_size = self._alloc_chunk(_sizes, data_flags)
                    if chunk_size == 0:
                        return _sizes, virtual_metadata, virtual_data
                    virtual_data += chunk_size
        else:
            flags = self.target_profile_mixed
            virtual_mixed = 0
            while True:
                chunk_size = self._alloc_chunk(_sizes, flags)
                if chunk_size == 0:
                    return _sizes, virtual_mixed
                virtual_mixed += chunk_size

    @staticmethod
    def _pretty_properties():
        return [
            (btrfs.utils.space_profile_description, 'target_profile_system'),
            (btrfs.utils.space_profile_description, 'target_profile_metadata'),
            (btrfs.utils.space_profile_description, 'target_profile_data'),
            (btrfs.utils.space_profile_description, 'target_profile_mixed'),
            (btrfs.utils.btrfs.utils.pretty_size, 'total'),
            (btrfs.utils.btrfs.utils.pretty_size, 'allocated'),
            (btrfs.utils.btrfs.utils.pretty_size, 'parity'),
            (btrfs.utils.btrfs.utils.pretty_size, 'virtual_total'),
            (btrfs.utils.btrfs.utils.pretty_size, 'virtual_used'),
            (btrfs.utils.btrfs.utils.pretty_size, 'unalloctable_soft'),
            (btrfs.utils.btrfs.utils.pretty_size, 'estimated_allocatable_virtual_metadata'),
            (btrfs.utils.btrfs.utils.pretty_size, 'estimated_allocatable_virtual_data'),
            (btrfs.utils.btrfs.utils.pretty_size, 'estimated_allocatable_virtual_mixed'),
            (btrfs.utils.btrfs.utils.pretty_size, 'unallocatable_hard'),
            (btrfs.utils.btrfs.utils.pretty_size, 'estimated_full_allocatable_virtual_metadata'),
            (btrfs.utils.btrfs.utils.pretty_size, 'estimated_full_allocatable_virtual_data'),
            (btrfs.utils.btrfs.utils.pretty_size, 'estimated_full_allocatable_virtual_mixed'),
            (btrfs.utils.btrfs.utils.pretty_size, 'unallocatable_reclaimable'),
            (btrfs.utils.btrfs.utils.pretty_size, 'allocatable'),
            (btrfs.utils.btrfs.utils.pretty_size, 'allocatable_left'),
            (btrfs.utils.btrfs.utils.pretty_size, 'free_metadata'),
            (btrfs.utils.btrfs.utils.pretty_size, 'free_data'),
            (btrfs.utils.btrfs.utils.pretty_size, 'free_mixed'),
        ]



# Copyright (C) 2016 Hans van Kranenburg <hans@knorrie.org>
#
# This file is part of the python-btrfs module.
#
# python-btrfs is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-btrfs is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-btrfs.  If not, see <http://www.gnu.org/licenses/>.

"""
This module contains miscellaneous collection of things, like the
:func:`pretty_print` function which provides a quick readable textual dump of
most of the object types in this library. Besides that, some functions to
pretty print and parse size strings and some other stuff.
"""

import btrfs
import collections.abc
import re
import types
from btrfs.ctree import (
    BLOCK_GROUP_DATA, BLOCK_GROUP_SYSTEM, BLOCK_GROUP_METADATA,
    SPACE_INFO_GLOBAL_RSV, BLOCK_GROUP_TYPE_MASK,
    BLOCK_GROUP_RAID0, BLOCK_GROUP_RAID1, BLOCK_GROUP_RAID5,
    BLOCK_GROUP_RAID1C3, BLOCK_GROUP_RAID1C4,
    BLOCK_GROUP_RAID6, BLOCK_GROUP_DUP, BLOCK_GROUP_RAID10,
    BLOCK_GROUP_SINGLE,
    BLOCK_GROUP_PROFILE_MASK,
)

SZ_1K = 0x00000400
SZ_2K = 0x00000800
SZ_4K = 0x00001000
SZ_8K = 0x00002000
SZ_16K = 0x00004000
SZ_32K = 0x00008000
SZ_64K = 0x00010000
SZ_128K = 0x00020000
SZ_256K = 0x00040000
SZ_512K = 0x00080000

SZ_1M = 0x00100000
SZ_2M = 0x00200000
SZ_4M = 0x00400000
SZ_8M = 0x00800000
SZ_16M = 0x01000000
SZ_32M = 0x02000000
SZ_64M = 0x04000000
SZ_128M = 0x08000000
SZ_256M = 0x10000000
SZ_512M = 0x20000000

SZ_1G = 0x40000000


def mounted_filesystem_paths():
    """Discover mountpoints for online btrfs filesystems

    :returns: Filesystem paths where btrfs filesystems are mounted.
    :rtype: List[str]
    """
    filesystems = {}
    mounts = [line.split() for line in open('/proc/self/mounts', 'r').read().splitlines()]
    for path in [mount[1] for mount in mounts if mount[2] == 'btrfs']:
        with btrfs.ctree.FileSystem(path) as fs:
            filesystems.setdefault(fs.fsid, path)
    return list(filesystems.values())


_space_type_description_map = {
    BLOCK_GROUP_DATA: 'Data',
    BLOCK_GROUP_SYSTEM: 'System',
    BLOCK_GROUP_METADATA: 'Metadata',
    BLOCK_GROUP_DATA | BLOCK_GROUP_METADATA: 'Data+Metadata',
    SPACE_INFO_GLOBAL_RSV: 'GlobalReserve',
}


def space_type_description(flags):
    """
    :param int flags: Space flags.
    :returns: String representation of the space type only.
    :rtype: str

    Example::

        >>> btrfs.utils.space_type_description(0x14)
        'Metadata'
    """
    return _space_type_description_map.get(
        flags & (BLOCK_GROUP_TYPE_MASK | SPACE_INFO_GLOBAL_RSV),
        'unknown'
    )


_space_profile_description_map = {
    BLOCK_GROUP_SINGLE: 'single',
    BLOCK_GROUP_RAID0: 'RAID0',
    BLOCK_GROUP_RAID1: 'RAID1',
    BLOCK_GROUP_RAID1C3: 'RAID1C3',
    BLOCK_GROUP_RAID1C4: 'RAID1C4',
    BLOCK_GROUP_RAID5: 'RAID5',
    BLOCK_GROUP_RAID6: 'RAID6',
    BLOCK_GROUP_DUP: 'DUP',
    BLOCK_GROUP_RAID10: 'RAID10',
}


def space_profile_description(flags):
    """
    :param int flags: Space flags.
    :returns: String representation of the space profile only.
    :rtype: str

    Example::

        >>> btrfs.utils.space_profile_description(0x14)
        'RAID1'
    """
    return _space_profile_description_map.get(
        flags & BLOCK_GROUP_PROFILE_MASK,
        'unknown'
    )


def space_flags_description(flags):
    """
    :param int flags: Space flags.
    :returns: String representation of the space type and profile.
    :rtype: str

    Example::

        >>> btrfs.utils.space_flags_description(0x14)
        'Metadata, RAID1'
    """
    return "{}, {}".format(
        btrfs.utils.space_type_description(flags),
        btrfs.utils.space_profile_description(flags))


pretty_size_units = '_KMGTPE'


def pretty_size(size, unit=None, binary=True):
    """
    :param int size: Size in bytes.
    :param str unit: Target unit to display the size in. One of 'kMGTPE'.
    :param bool binary: If True, use base 1024, else base 1000.

    Example::

        >>> btrfs.utils.pretty_size(1610612736, unit='G')
        '1.50GiB'
    """
    if unit == '':
        return str(size)
    base = 1024 if binary else 1000
    if unit is None:
        if size == 0:
            unit = ''
            unit_offset = 0
            base = 1000
        else:
            unit_offset = 0
            tmp = size
            while tmp >= 1024 and unit_offset < len(pretty_size_units) - 1:
                unit_offset = unit_offset + 1
                tmp = tmp / 1024
            unit = pretty_size_units[unit_offset] if unit_offset > 0 else ''
    else:
        unit = unit.upper()
        unit_offset = pretty_size_units.index(unit)
        if unit == 'K' and base == 1000:
            unit = 'k'
    divide_by = base ** unit_offset
    if divide_by > 0:
        size = size / divide_by
    return "{0:.2f}{1}{2}B".format(size, unit, 'i' if base == 1024 and unit != '' else '')


_re_parse_pretty_size = re.compile(r'^(?P<size>\d+)(?:(?P<unit>[kMGTPE])((?P<i>i)?B)?)?$')


def parse_pretty_size(size_str):
    """Parse pretty printed size strings.

    This is the opposite of the :func:`pretty_size` function, but not 100%.

    :param str size_str: String literal to be parsed.
    :raises ValueError: If the input cannot be parsed as pretty size.

    Example::

            >>> btrfs.utils.parse_pretty_size('234MB')
            234000000
            >>> btrfs.utils.parse_pretty_size('234MiB')
            245366784
            >>> btrfs.utils.parse_pretty_size('234TB')
            234000000000000
            >>> btrfs.utils.parse_pretty_size('234TiB')
            257285720899584
            >>> btrfs.utils.parse_pretty_size('1048576')
            1048576
    """
    match = _re_parse_pretty_size.match(size_str)
    if match is None:
        raise ValueError('literal cannot be parsed as pretty size')
    groupdict = match.groupdict()
    if groupdict['unit'] is None:
        return int(size_str)
    base = 1024 if groupdict['i'] == 'i' else 1000
    unit_offset = pretty_size_units.index(groupdict['unit'].upper())
    multiply_by = base ** unit_offset
    return int(groupdict['size']) * multiply_by


def flags_str(flags, flags_str_map):
    """Generic helper to convert flags to a description.

    This function is used by more specific helper functions below. You probably
    don't need to call it directly.

    :param int flags: Flags.
    :param flags_str_map: Mapping from flag bit value to description.
    :type flags_str_map: dict(int, str)
    """
    ret = []
    for flag in sorted(flags_str_map.keys()):
        if flags & flag:
            ret.append(flags_str_map[flag])
            flags ^= flag
    if flags != 0:
        ret.append("unknown(0x{:0x})".format(flags))
    elif len(ret) == 0:
        ret.append("none")
    return '|'.join(ret)


def block_group_flags_str(flags):
    """
    :param int flags: Block Group flags.
    :returns: String representation of the block group type and profile.
    :rtype: str

    Example::

        >>> btrfs.utils.block_group_flags_str(0x14)
        'METADATA|RAID1'
    """
    return flags_str(flags, btrfs.ctree._block_group_flags_str_map)


def block_group_type_str(flags):
    """
    :param int flags: Block Group flags.
    :returns: String representation of the block group type only.
    :rtype: str

    Example::

        >>> btrfs.utils.block_group_type_str(0x14)
        'METADATA'
    """
    return block_group_flags_str(flags & BLOCK_GROUP_TYPE_MASK)


def block_group_profile_str(flags):
    """
    :param int flags: Block Group flags.
    :returns: String representation of the block group profile only.
    :rtype: str

    Example::

        >>> btrfs.utils.block_group_profile_str(0x14)
        'RAID1'
    """
    return block_group_flags_str(flags & BLOCK_GROUP_PROFILE_MASK)


def extent_flags_str(flags):
    """
    :param int flags: :class:`~btrfs.ctree.ExtentItem` flags.
    :returns: String representation of the extent item flags.
    :rtype: str

    Example::

        >>> btrfs.utils.extent_flags_str(0x102)
        'TREE_BLOCK|FULL_BACKREF'
    """
    return flags_str(flags, btrfs.ctree._extent_flags_str_map)


def inode_mode_str(mode):
    """
    :param int mode: File mode as number.
    :returns: String representation of file mode in octal format.
    :rtype: str

    Example::

        >>> btrfs.utils.inode_mode_str(16877)
        '040755'

    """
    return "0{:05o}".format(mode)


def inode_flags_str(flags):
    """
    :param int flags: :class:`~btrfs.ctree.InodeItem` flags.
    :returns: String representation of the inode item flags.
    :rytpe: str

    Example::

        >>> btrfs.utils.inode_flags_str(72)
        'NOCOMPRESS|IMMUTABLE'
    """
    return flags_str(flags, btrfs.ctree._inode_flags_str_map)


def dir_item_type_str(type_):
    """
    :param int type_: :class:`~btrfs.ctree.DirItem` type.
    :returns: String representation of DirItem type.
    :rtype: str

    Example::

        >>> btrfs.utils.dir_item_type_str(4)
        'BLKDEV'
    """
    return btrfs.ctree._dir_item_type_str_map[type_]


def root_item_flags_str(flags):
    """
    :param int flags: :class:`~btrfs.ctree.RootItem` flags.
    :returns: String representation of RootItem flags.
    :rtype: str

    Example::

        >>> btrfs.utils.root_item_flags_str(1)
        'RDONLY'
    """
    return flags_str(flags, btrfs.ctree._root_flags_str_map)


def compress_type_str(compression):
    """
    :param int compression: :class:`~btrfs.ctree.FileExtentItem` compression
        type.
    :returns: String representation of compression type.
    :rtype: str

    Example::

        >>> btrfs.utils.compress_type_str(3)
        'zstd'
    """
    return btrfs.ctree._compress_type_str_map.get(compression, 'unknown')


def file_extent_type_str(type_):
    """
    :param int type_: :class:`~btrfs.ctree.FileExtentItem` type.
    :returns: String representation of FileExtentItem type.
    :rtype: str

    Example::

        >>> btrfs.utils.file_extent_type_str(0)
        'inline'
    """
    return btrfs.ctree._file_extent_type_str_map.get(type_, 'unknown')


def free_space_info_flags_str(flags):
    """
    :param int flags: :class:`~btrfs.ctree.FreeSpaceInfo` flags.
    :returns: String representation of FreeSpaceInfo flags.
    :rtype: str

    Example::

        >>> btrfs.utils.free_space_info_flags_str(1)
        'bitmaps'
    """
    return flags_str(flags, btrfs.ctree._free_space_info_flags_str_map)


def embedded_text_for_str(text):
    """
    :param bytes text: bytes from a name of data field of an object from the
        :class:`btrfs.ctree.DirItem` family.
    :returns: String representation of the bytes.
    :rtype: str

    This function is not intended to be used for anything else than a
    convenient way of displaying filenames and xattr keys and values in the
    pretty printer.

    Example::

        >>> name_bytes = b'\\xce\\xbc\\xce\\xbf\\xcf\\x85\\xcf\\x84\\xce\\xbf\\xce\\xbd'
        >>> btrfs.utils.embedded_text_for_str(name_bytes)
        "utf-8 'μουτον'"
    """
    if len(text) == 0:
        return "<empty>"
    try:
        return "utf-8 '{}'".format(text.decode('utf-8'))
    except UnicodeDecodeError:
        return "raw '{}'".format(repr(text))


_tree_name_id_map = {
    'root': btrfs.ctree.ROOT_TREE_OBJECTID,
    'extent': btrfs.ctree.EXTENT_TREE_OBJECTID,
    'chunk': btrfs.ctree.CHUNK_TREE_OBJECTID,
    'dev': btrfs.ctree.DEV_TREE_OBJECTID,
    'fs': btrfs.ctree.FS_TREE_OBJECTID,
    'csum': btrfs.ctree.CSUM_TREE_OBJECTID,
    'quota': btrfs.ctree.QUOTA_TREE_OBJECTID,
    'uuid': btrfs.ctree.UUID_TREE_OBJECTID,
    'free_space': btrfs.ctree.FREE_SPACE_TREE_OBJECTID,
    'block_group': btrfs.ctree.BLOCK_GROUP_TREE_OBJECTID,
    'tree_log': btrfs.ctree.TREE_LOG_OBJECTID,
    'tree_log_fixup': btrfs.ctree.TREE_LOG_FIXUP_OBJECTID,
    'tree_reloc': btrfs.ctree.TREE_RELOC_OBJECTID,
    'data_reloc': btrfs.ctree.DATA_RELOC_TREE_OBJECTID,
}


def parse_tree_name(name):
    """Convert a tree name to an actual root tree objectid number

    :param string name: tree name with optional _tree suffix or number
        formatted as string
    :returns: Tree objectid that is valid for the root tree
    :rtype: int

    Example::

        >>> btrfs.utils.parse_tree_name('EXTENT_TREE')
        2
        >>> btrfs.utils.parse_tree_name('extent_tree')
        2
        >>> btrfs.utils.parse_tree_name('extent')
        2
        >>> btrfs.utils.parse_tree_name('2')
        2
        >>> btrfs.utils.parse_tree_name(2)
        2
    """
    try:
        return int(name)
    except ValueError:
        pass
    lower_name = name.lower()
    if lower_name[-5:] == '_tree':
        lookup_name = lower_name[:-5]
    else:
        lookup_name = lower_name
    if lookup_name in _tree_name_id_map:
        return _tree_name_id_map[lookup_name]
    raise ValueError("Unknown metadata tree name " + name)


def parse_key_string(key_str):
    """Create a Key object from a pretty printed string representation

    :param string key_str: String representation of a key, e.g. '(31832 INODE_REF 31798)'
    :returns: A Key object with the parsed values set
    :rtype: :class:`~btrfs.ctree.Key`
    :raises: :class:`ValueError` if the key string is an invalid key representation

    The parentheses around the key triplet are optional.

    Example::

        >>> btrfs.utils.parse_key_string('(31832 INODE_REF 31798)')
        Key(31832, 12, 31798)
        >>> btrfs.utils.parse_key_string('(535 EXTENT_DATA 0)')
        Key(535, 108, 0)

    """
    if key_str[0] == '(' and key_str[-1] == ')':
        key_str = key_str[1:-1]
    try:
        objectid_str, type_str, offset_str = key_str.split()
    except ValueError:
        raise ValueError(
            "Key representation needs 3 fields: objectid, type, offset: {}".format(key_str)) \
            from None
    return btrfs.ctree.Key(objectid_str, type_str, offset_str)


def _pretty_attr_value(obj, attr_name, stringify_fn=None):
    if stringify_fn is not None:
        return "{}: {}".format(attr_name, stringify_fn(getattr(obj, attr_name)))
    cls = obj.__class__
    attr_name_str = '_{}_str'.format(attr_name)
    stringify_property = getattr(cls, attr_name_str, None)
    if stringify_property is None:
        # try deprecated pattern
        attr_name_str = '{}_str'.format(attr_name)
        stringify_property = getattr(cls, attr_name_str, None)
    if stringify_property is not None and isinstance(stringify_property, property):
        return "{}: {}".format(attr_name, getattr(obj, attr_name_str))
    value = getattr(obj, attr_name)
    if isinstance(value, (memoryview, bytes)):
        return "{}: {} bytes of data".format(attr_name, len(value))
    return "{}: {}".format(attr_name, value)


pretty_print_modules = 'btrfs.ctree', 'btrfs.ioctl', 'btrfs.fs_usage', 'btrfs.free_space_tree'


def _pretty_obj_tuples(obj, level=0, seen=None):
    if seen is None:
        seen = []
    cls = obj.__class__
    if isinstance(obj, btrfs.ctree.ItemData) and \
            hasattr(obj, 'key') and isinstance(obj.key, btrfs.ctree.Key):
        yield level, "<{}.{} {}>".format(cls.__module__, cls.__name__, str(obj.key))
    elif cls.__module__ in pretty_print_modules:
        yield level, "<{}.{}>".format(cls.__module__, cls.__name__)
    if obj in seen:
        yield level, "[... object already seen, aborting recursion]"
        return
    seen.append(obj)
    known = False
    if cls.__module__ in pretty_print_modules and \
            not isinstance(obj, btrfs.ctree.Key):
        known = True
        if isinstance(obj, btrfs.ctree.ItemData):
            try:
                objectid_attr, type_attr, offset_attr = obj._key_attrs
                if objectid_attr is not None:
                    yield level, "{} (key objectid)".format(_pretty_attr_value(obj, objectid_attr))
                if type_attr is not None:
                    yield level, "{} (key type)".format(_pretty_attr_value(obj, type_attr))
                if offset_attr is not None:
                    yield level, "{} (key offset)".format(_pretty_attr_value(obj, offset_attr))
            except AttributeError:
                pass
        for attr_name, attr_value in obj.__dict__.items():
            if attr_name.startswith('_'):
                continue
            if isinstance(obj, btrfs.ctree.ItemData):
                try:
                    if attr_name in obj._key_attrs:
                        continue
                except AttributeError:
                    pass
                if attr_name == 'key' and isinstance(attr_value, btrfs.ctree.Key):
                    continue
            if isinstance(attr_value, list):
                if len(attr_value) == 0:
                    continue
                yield level, "{}:".format(attr_name)
                for item in attr_value:
                    yield level, '-'
                    yield from _pretty_obj_tuples(item, level+1, seen)
            elif isinstance(attr_value, dict):
                if len(attr_value) == 0:
                    continue
                yield level, "{}:".format(attr_name)
                for k, v in attr_value.items():
                    dict_key_str = '_{}_key_str'.format(attr_name)
                    stringify_fn = getattr(cls, dict_key_str, None)
                    if stringify_fn is None:
                        # try deprecated pattern
                        dict_key_str = '{}_key_str'.format(attr_name)
                        stringify_fn = getattr(cls, dict_key_str, None)
                    if stringify_fn is not None and callable(stringify_fn):
                        yield level+1, "{}:".format(stringify_fn(k))
                    else:
                        yield level+1, "{}:".format(k)
                    yield from _pretty_obj_tuples(v, level+2, seen)
            elif attr_value.__class__.__module__ in pretty_print_modules and \
                    not isinstance(attr_value, (btrfs.ctree.Key, btrfs.ctree.TimeSpec)):
                yield level, "{}:".format(attr_name)
                yield from _pretty_obj_tuples(attr_value, level+1, seen)
            else:
                yield level, _pretty_attr_value(obj, attr_name)
    if isinstance(obj, (list, types.GeneratorType)) or \
            (isinstance(obj, btrfs.ctree.ItemData) and
             isinstance(obj, collections.abc.MutableSequence)):
        known = True
        for item in obj:
            yield level, '-'
            yield from _pretty_obj_tuples(item, level+1, seen)
    elif isinstance(obj, dict) and len(obj) != 0:
        known = True
        for k, v in obj.items():
            yield level, "{}:".format(k)
            yield from _pretty_obj_tuples(v, level+1, seen)
    if not known:
        yield level, str(obj)
    seen.pop()


def _pretty_obj_lines(obj, level=0):
    for level, line in _pretty_obj_tuples(obj, level):
        yield "{}{}".format('  ' * level, line)


def _pretty_print(obj, level=0):
    for line in _pretty_obj_lines(obj, level):
        print(line)


def pretty_print(obj):
    """
    The pretty printer dumps content of objects on the screen. It is mainly
    designed to work with objects from the btrfs library. It is also possible
    to provide a lists of objects, or a generator, or even nested structures.

    :param obj: An object to pretty print, or a collection of them.
    :type obj: anything goes

    Example::

        >>> import btrfs
        >>> with btrfs.FileSystem('/') as fs:
        ...     btrfs.utils.pretty_print(fs.block_group(1687427219456))
        ...
        <btrfs.ctree.BlockGroupItem (1687427219456 BLOCK_GROUP_ITEM 1073741824)>
        vaddr: 1687427219456 (key objectid)
        length: 1.00GiB (key offset)
        flags: DATA
        chunk_objectid: 256
        used: 960.50MiB
    """
    _pretty_print(obj)


def _str_obj_tuples(obj, level=0, seen=None):
    if seen is None:
        seen = []
    cls = obj.__class__
    if cls.__module__ in pretty_print_modules and \
            not isinstance(obj, (btrfs.ctree.Key, btrfs.ctree.TimeSpec)):
        yield level, str(obj)
    if obj in seen:
        yield level, "[... object already seen, aborting recursion]"
        return
    seen.append(obj)
    if isinstance(obj, btrfs.ctree.ItemData):
        for attr_name, attr_value in obj.__dict__.items():
            if attr_name.startswith('_'):
                continue
            if isinstance(obj, (btrfs.ctree.ItemData, btrfs.ctree.SubItem)):
                yield from _str_obj_tuples(attr_value, level+1, seen)
            elif isinstance(attr_value, list):
                for item in attr_value:
                    yield from _str_obj_tuples(item, level+1, seen)
    if isinstance(obj, (list, types.GeneratorType)):
        for item in obj:
            yield from _str_obj_tuples(item, level, seen)
    elif isinstance(obj, btrfs.ctree.ItemData) and \
            isinstance(obj, collections.abc.MutableSequence):
        for item in obj:
            yield from _str_obj_tuples(item, level+1, seen)
    seen.pop()


def _str_print_lines(obj, level=0):
    for level, line in _str_obj_tuples(obj, level):
        yield "{}{}".format('  ' * level, line)


def str_print(obj):
    """
    Print the usual str() of an object, but look inside to see if more ItemData
    objects are hidden there. If so, also print their str().

    :param obj: An object to pretty print, or a collection of them.
    :type obj: anything goes

    Example::

        >>> with btrfs.FileSystem('/') as fs:
        ...      btrfs.utils.str_print(fs.block_group(63381176320))
        ...
        block group vaddr 63381176320 length 1073741824 flags DATA used 1073741824 used_pct 100
    """
    for line in _str_print_lines(obj):
        print(line)



# Copyright (C) 2018 Hans van Kranenburg <hans@knorrie.org>
#
# This file is part of the python-btrfs module.
#
# python-btrfs is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# python-btrfs is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with python-btrfs.  If not, see <http://www.gnu.org/licenses/>.


from btrfs.ctree import (  # noqa
    BLOCK_GROUP_DATA, BLOCK_GROUP_SYSTEM, BLOCK_GROUP_METADATA,
    SPACE_INFO_GLOBAL_RSV, BLOCK_GROUP_TYPE_MASK,
    BLOCK_GROUP_RAID0, BLOCK_GROUP_RAID1, BLOCK_GROUP_RAID5,
    BLOCK_GROUP_RAID1C3, BLOCK_GROUP_RAID1C4,
    BLOCK_GROUP_RAID6, BLOCK_GROUP_DUP, BLOCK_GROUP_RAID10,
    BLOCK_GROUP_SINGLE,
    BLOCK_GROUP_PROFILE_MASK,
)
from btrfs.utils import SZ_1G
from collections import namedtuple

BTRFS_MAX_DATA_CHUNK_SIZE = 10 * SZ_1G

_RaidAttr = namedtuple('RaidAttr', [
    'sub_stripes', 'dev_stripes', 'devs_max', 'devs_min', 'tolerated_failures',
    'devs_increment', 'ncopies', 'nparity', 'raid_name', 'bg_flag',
])

RAID_RAID10 = 0
RAID_RAID1 = 1
RAID_RAID1C3 = 2
RAID_RAID1C4 = 3
RAID_DUP = 4
RAID_RAID0 = 5
RAID_SINGLE = 6
RAID_RAID5 = 7
RAID_RAID6 = 8


def _bg_flags_to_raid_index(flags):
    """Convert block group flags to an index to access _raid_array"""
    if flags & BLOCK_GROUP_RAID10:
        return RAID_RAID10
    if flags & BLOCK_GROUP_RAID1:
        return RAID_RAID1
    if flags & BLOCK_GROUP_RAID1C3:
        return RAID_RAID1C3
    if flags & BLOCK_GROUP_RAID1C4:
        return RAID_RAID1C4
    if flags & BLOCK_GROUP_DUP:
        return RAID_DUP
    if flags & BLOCK_GROUP_RAID0:
        return RAID_RAID0
    if flags & BLOCK_GROUP_RAID5:
        return RAID_RAID5
    if flags & BLOCK_GROUP_RAID6:
        return RAID_RAID6
    return RAID_SINGLE


_raid_array = [
    _RaidAttr(
        sub_stripes=2,
        dev_stripes=1,
        devs_max=0,
        devs_min=4,
        tolerated_failures=1,
        devs_increment=2,
        ncopies=2,
        nparity=0,
        raid_name='raid10',
        bg_flag=BLOCK_GROUP_RAID10,
    ),
    _RaidAttr(
        sub_stripes=1,
        dev_stripes=1,
        devs_max=2,
        devs_min=2,
        tolerated_failures=1,
        devs_increment=2,
        ncopies=2,
        nparity=0,
        raid_name='raid1',
        bg_flag=BLOCK_GROUP_RAID1,
    ),
    _RaidAttr(
        sub_stripes=1,
        dev_stripes=1,
        devs_max=3,
        devs_min=3,
        tolerated_failures=2,
        devs_increment=3,
        ncopies=3,
        nparity=0,
        raid_name='raid1c3',
        bg_flag=BLOCK_GROUP_RAID1C3,
    ),
    _RaidAttr(
        sub_stripes=1,
        dev_stripes=1,
        devs_max=4,
        devs_min=4,
        tolerated_failures=3,
        devs_increment=4,
        ncopies=4,
        nparity=0,
        raid_name='raid1c4',
        bg_flag=BLOCK_GROUP_RAID1C4,
    ),
    _RaidAttr(
        sub_stripes=1,
        dev_stripes=2,
        devs_max=1,
        devs_min=1,
        tolerated_failures=0,
        devs_increment=1,
        ncopies=2,
        nparity=0,
        raid_name='dup',
        bg_flag=BLOCK_GROUP_DUP,
    ),
    _RaidAttr(
        sub_stripes=1,
        dev_stripes=1,
        devs_max=0,
        devs_min=2,
        tolerated_failures=0,
        devs_increment=1,
        ncopies=1,
        nparity=0,
        raid_name='raid0',
        bg_flag=BLOCK_GROUP_RAID0,
    ),
    _RaidAttr(
        sub_stripes=1,
        dev_stripes=1,
        devs_max=1,
        devs_min=1,
        tolerated_failures=0,
        devs_increment=1,
        ncopies=1,
        nparity=0,
        raid_name='single',
        bg_flag=BLOCK_GROUP_SINGLE,
    ),
    _RaidAttr(
        sub_stripes=1,
        dev_stripes=1,
        devs_max=0,
        devs_min=2,
        tolerated_failures=1,
        devs_increment=1,
        ncopies=1,
        nparity=1,
        raid_name='raid5',
        bg_flag=BLOCK_GROUP_RAID5,
    ),
    _RaidAttr(
        sub_stripes=1,
        dev_stripes=1,
        devs_max=0,
        devs_min=3,
        tolerated_failures=2,
        devs_increment=1,
        ncopies=1,
        nparity=2,
        raid_name='raid6',
        bg_flag=BLOCK_GROUP_RAID6,
    ),
]


def _raid_attrs(flags):
    return _raid_array[_bg_flags_to_raid_index(flags)]


def chunk_length_to_dev_extent_length(flags, num_stripes, chunk_length):
    """Given information about a :class:`~btrfs.ctree.Chunk`, calculate the
    length of :class:`~btrfs.ctree.DevExtent` objects that store the Chunk
    data.

    For example, a Data, RAID6 block group of 8GiB, distributed over 6 devices,
    will occupy 2GiB * (4+2) = 12 GiB physical allocated bytes since it has
    4GiB of allocated bytes reserved for parity:

    Example::

        >>> btrfs.utils.pretty_size(
        ...     btrfs.volumes.chunk_length_to_dev_extent_length(
        ...         btrfs.ctree.BLOCK_GROUP_RAID6, 6, 8 * btrfs.utils.SZ_1G))
        '2.00GiB'
    """
    # So, we start with a chunk length, which is the amount of usable virtual
    # space.
    attrs = _raid_attrs(flags & BLOCK_GROUP_PROFILE_MASK)
    # The nparity attribute means that we have nparity * dev_extent_length of
    # raw space in total, dedicated for parity. These parity bytes can be
    # distributed over all of the num_stripes. If we subtract nparity from
    # num_stripes, we get the amount of dev_extent_lengths worth that contain
    # only data.
    #
    # For simplicity, let's assume all parity lives within dedicated device
    # extents. In reality it robins around, but that doesn't matter for our
    # calculations.
    #
    # RAID5: D | D | D | P
    # RAID6: D | D | D | P | P
    num_data_stripes = num_stripes - attrs.nparity
    # In case of profiles that duplicate data, we have to correct for that.
    # So, if we multiply the chunk_length (which is virtual space) by the
    # amount of copies of the data, we get the amount of raw bytes that we have
    # to fit in num_data_stripes amount of device extents.
    raw_data_bytes = chunk_length * attrs.ncopies
    dev_extent_length = raw_data_bytes // num_data_stripes
    return dev_extent_length


def chunk_to_dev_extent_length(chunk):
    """Given a :class:`~btrfs.ctree.Chunk` object, calculate the length of
    :class:`~btrfs.ctree.DevExtent` objects that store the chunk data.

    Example::

        >>> with btrfs.FileSystem('/') as fs:
        ...     chunk = list(fs.chunks(min_vaddr=3250585600, nr_items=1))[0]
        ...     print(chunk)
        ...     dev_extent_length = btrfs.volumes.chunk_to_dev_extent_length(chunk)
        ...     print("device extent length is {}".format(
        ...         btrfs.utils.pretty_size(dev_extent_length)))
        ...
        chunk vaddr 3250585600 type DATA|RAID5 length 2147483648 num_stripes 3
        device extent length is 1.00GiB
    """
    return chunk_length_to_dev_extent_length(chunk.type, chunk.num_stripes, chunk.length)


def dev_extent_length_to_chunk_length(flags, num_stripes, stripe_size):
    """This function simply reverses the calculation of
    :class:`~btrfs.ctree.Chunk` length to :class:`~btrfs.ctree.DevExtent`
    length.

    Example::

        >>> btrfs.utils.pretty_size(
        ...     btrfs.volumes.dev_extent_length_to_chunk_length(
        ...         btrfs.ctree.BLOCK_GROUP_RAID6, 6, 2 * btrfs.utils.SZ_1G))
        '8.00GiB'
    """
    attrs = _raid_attrs(flags & BLOCK_GROUP_PROFILE_MASK)
    num_data_stripes = num_stripes - attrs.nparity
    # stripe_size is a synonym for device extent length.
    raw_data_bytes = stripe_size * num_data_stripes
    chunk_length = raw_data_bytes // attrs.ncopies
    return chunk_length


def chunk_to_raw_parity_bytes(chunk):
    """Given a :class:`~btrfs.ctree.Chunk` object, calculate how many bytes are
    reserved for storing parity data for RAID56 profiles.

    This number is relevant to understand how much raw disk space that is
    allocated but not used is actually not usable for data, because it's
    reserved for parity.

    Example::

        >>> with btrfs.FileSystem('/mnt/tutorial') as fs:
        ...  chunk = list(fs.chunks(min_vaddr=3250585600, nr_items=1))[0]
        ...  print(chunk)
        ...  parity_bytes = btrfs.volumes.chunk_to_raw_parity_bytes(chunk)
        ...  print("amount of raw parity bytes: {}".format(
        ...      btrfs.utils.pretty_size(parity_bytes)))
        ...
        chunk vaddr 3250585600 type DATA|RAID5 length 2147483648 num_stripes 3
        amount of raw parity bytes: 1.00GiB
    """
    dev_extent_length = chunk_to_dev_extent_length(chunk)
    attrs = _raid_attrs(chunk.type & BLOCK_GROUP_PROFILE_MASK)
    return dev_extent_length * attrs.nparity


def block_group_profile_ncopies(flags):
    """This function returns how many times the actual data is replicated on
    disk given block group flags as input. E.g. for RAID1 this is 2, but for
    RAID5, this is just 1. The actual data is stored once in case of RAID5, and
    the redundancy is done using parity blocks, not data itself.

    Example::

        >>> btrfs.volumes.block_group_profile_ncopies(btrfs.ctree.BLOCK_GROUP_RAID10)
        2
    """
    attrs = _raid_attrs(flags & BLOCK_GROUP_PROFILE_MASK)
    return attrs.ncopies
